#include <cppunit/TestFixture.h>
#include <cppunit/extensions/HelperMacros.h>
#include "MyAssertionTraits.h"

#include "Configuration.h"
#include "XMLVisitors.h"

#include "Tiles.h"
#include "PlayerConfiguration.h"

#include "AllObjects.h"
#include "ObjectRepository.h"
#include "GameBridge.h"
#include "GameControlBase.h"


//----------------------------------------------------------------------------
class GameControlTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(GameControlTest);

    CPPUNIT_TEST(testLandObjectives);
    CPPUNIT_TEST(testDestroyObjectives);
    CPPUNIT_TEST(testRescueObjectives);

    CPPUNIT_TEST_SUITE_END();

public:

    //------------------------------------------------------------------------
    void setUp()
    {
        Configuration::getInstance()->searchDataDir();

        SDLFrameRate::setFrameRate(72);

        TileCategories::init();
        SurfacesBase::init();

        PlayerConfiguration::create();

        GameInterface::setInstance(GameBridge::getInstance());
    }

    //------------------------------------------------------------------------
    void tearDown() 
    {
        PlayerConfiguration::destroy();

        SurfacesBase::destroy();
        TileCategories::destroy();
    }


protected:

    //------------------------------------------------------------------------
    void testLandObjectives()
    {
        GameControlBase::init("test", "land.xml");
        GameControlBase *gc = GameControlBase::getInstance();
        Ship *ship = gc->getPlayerShip();

        gc->onShipLanded(ship, getPlatform(1));
        assertObjectives(
            "<land platform=\"2,3\"/>\n"
            "<land platform=\"4,5\"/>\n");

        gc->onShipLanded(ship, getPlatform(2));
        assertObjectives(
            "<land platform=\"3\"/>\n"
            "<land platform=\"4,5\"/>\n");

        gc->onShipLanded(ship, getPlatform(3));
        assertObjectives(
            "<land platform=\"4,5\"/>\n");

        gc->onShipLanded(ship, getPlatform(4));
        assertObjectives(
            "<land platform=\"5\"/>\n");

        gc->onShipLanded(ship, getPlatform(5));
        assertObjectives(
            "");

        CPPUNIT_ASSERT_EQUAL(
            (const char*)"AddFuelToScore", gc->getNameOfCurrentState());

        GameControlBase::destroy();
    }

    //------------------------------------------------------------------------
    void testDestroyObjectives()
    {
        GameControlBase::init("test", "destroy.xml");
        GameControlBase *gc = GameControlBase::getInstance();
        Ship *ship = gc->getPlayerShip();

        gc->onShipLanded(ship, getPlatform(2));
        assertObjectives(
            "<destroy id=\"10,11\"/>\n"
            "<destroy id=\"12,13\"/>\n"
            "<land platform=\"2\"/>\n");

        gc->onObjectDestroyed(getObject(13));
        assertObjectives(
            "<destroy id=\"10,11\"/>\n"
            "<destroy id=\"12\"/>\n"
            "<land platform=\"2\"/>\n");

        gc->onObjectDestroyed(getObject(10));
        assertObjectives(
            "<destroy id=\"11\"/>\n"
            "<destroy id=\"12\"/>\n"
            "<land platform=\"2\"/>\n");

        gc->onObjectDestroyed(getObject(12));
        assertObjectives(
            "<destroy id=\"11\"/>\n"
            "<land platform=\"2\"/>\n");

        gc->onObjectDestroyed(getObject(11));
        assertObjectives(
            "<land platform=\"2\"/>\n");

        gc->onShipLanded(ship, getPlatform(2));
        assertObjectives(
            "");

        CPPUNIT_ASSERT_EQUAL(
            (const char*)"AddFuelToScore", gc->getNameOfCurrentState());

        GameControlBase::destroy();
    }

    //------------------------------------------------------------------------
    void testRescueObjectives()
    {
        GameControlBase::init("test", "rescue.xml");
        GameControlBase *gc = GameControlBase::getInstance();
        Ship *ship = gc->getPlayerShip();

        movePlayerShipToCrate(10);
        gc->onShipLanded(ship, getPlatform(2)); // Pick up crate 10.
        assertObjectives(
            "<pickup id=\"11\" platform=\"5\"/>\n"
            "<pickup id=\"12\" platform=\"1\"/>\n");

        gc->onShipLanded(ship, getPlatform(1)); // Land on the wrong platform.
        assertObjectives(
            "<pickup id=\"11\" platform=\"5\"/>\n"
            "<pickup id=\"12\" platform=\"1\"/>\n");

        gc->onShipLanded(ship, getPlatform(5)); // Unload crate 10.
        unloadCrates(1);
        assertObjectives(
            "<pickup id=\"11\" platform=\"5\"/>\n"
            "<pickup id=\"12\" platform=\"1\"/>\n");

        movePlayerShipToCrate(12);
        gc->onShipLanded(ship, getPlatform(4)); // 12 can't be picked up yet.
        CPPUNIT_ASSERT(!ship->hasCrates());
        assertObjectives(
            "<pickup id=\"11\" platform=\"5\"/>\n"
            "<pickup id=\"12\" platform=\"1\"/>\n");

        movePlayerShipToCrate(11);
        gc->onShipLanded(ship, getPlatform(3)); // Pick up crate 11.
        assertObjectives(
            "<pickup id=\"\" platform=\"5\"/>\n"
            "<pickup id=\"12\" platform=\"1\"/>\n");

        gc->onShipLanded(ship, getPlatform(5)); // Unload crate 11.
        unloadCrates(1);
        assertObjectives(
            "<pickup id=\"12\" platform=\"1\"/>\n");
        
        movePlayerShipToCrate(12);
        gc->onShipLanded(ship, getPlatform(4)); // Pick up crate 12.
        assertObjectives(
            "<pickup id=\"\" platform=\"1\"/>\n");

        gc->onShipLanded(ship, getPlatform(1)); // Unload crate 12.
        unloadCrates(1);
        assertObjectives(
            "");

        CPPUNIT_ASSERT_EQUAL(
            (const char*)"AddFuelToScore", gc->getNameOfCurrentState());

        GameControlBase::destroy();
    }

private:

    //------------------------------------------------------------------------
    ObjectBase *getObject(unsigned id)
    {
        return ObjectRepository::getInstance()->getObject(id);
    }

    //------------------------------------------------------------------------
    Platform *getPlatform(unsigned id)
    {
        return dynamic_cast<Platform*>(
            ObjectRepository::getInstance()->getObject(id));
    }

    //------------------------------------------------------------------------
    void movePlayerShipToCrate(unsigned crateId)
    {
        Ship *ship = GameControlBase::getInstance()->getPlayerShip();
        const SDL_Rect &shipPos = ship->getPosition();

        Crate *crate = dynamic_cast<Crate*>(
            ObjectRepository::getInstance()->getObject(crateId));
        const SDL_Rect &cratePos = crate->getPosition();

        ship->setPosition(cratePos.x + cratePos.w/2 - shipPos.w/2,
                          cratePos.y + cratePos.h - shipPos.h);
    }

    //------------------------------------------------------------------------
    void unloadCrates(unsigned num)
    {
        GameControlBase *gc = GameControlBase::getInstance();

        for (unsigned i=num; i>0; i--)
        {
            CPPUNIT_ASSERT_EQUAL(
                (const char*)"UnloadingCrate",
                gc->getNameOfCurrentState());
            while (!strcmp(gc->getNameOfCurrentState(), "UnloadingCrate"))
            {
                gc->onPreUpdate();
            }

            if (i > 1)
            {
                CPPUNIT_ASSERT_EQUAL(
                    (const char*)"UnloadingDelay",
                    gc->getNameOfCurrentState());
                while (!strcmp(gc->getNameOfCurrentState(), "UnloadingDelay"))
                {
                    gc->onPreUpdate();
                }
            }
        }
    }

    //------------------------------------------------------------------------
    void assertObjectives(const char *desired)
    {
        std::ostringstream s;
        XMLWriteToStreamVisitor v(s);
        GameControlBase::getInstance()->getGameControlNode()->accept(v);

        CPPUNIT_ASSERT_EQUAL(std::string(desired), s.str());
    }
};

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(GameControlTest, "GameControl");
